/***************************************************************************
 *   Copyright (C) 2004 by Michael Moritz                                  *
 *   mimo@restoel.net                                                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <pthread.h>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <syslog.h>
#include "defs.h"
#include "triplet.h"
#include "read.h"
#include "configreader.h"
#include "cfg.h"
#include "db.h"
#include "action.h"
using namespace std;

const static char s_gpsname[] = PACKAGE; //from config.h

typedef out_of_range EndOfInputException;


void usage()
{
	cerr << "usage: gps [-v] configfile" << endl;
}
int end_usage(const std::string& msg)
{
	syslog(LOG_ERR, "Error: %s",msg.c_str());
	cerr << "Error: " << msg << "\n" << endl;
	usage();
	closelog();
	return -1;
}
/*
void action(const std::string& act)
{
	if(g_getCfg().isVerbose())
		syslog(LOG_DEBUG,"action=%s",act.c_str());
	cout << "action=" << act << "\n" << endl;
}
*/
#if 0
void *watchdog_runner(void *);

class Watchdog 
{
public:
	Watchdog() : _tid(0)
	{
		time(&_etime); 
		// set the default timeout to 30 secs
		_etime += 30;
  		/// get the default attributes 
  		pthread_attr_init(&_attr);
		pthread_attr_setdetachstate(&_attr,PTHREAD_CREATE_DETACHED);
  		/// create the thread 
  		pthread_create(&_tid, &_attr, watchdog_runner, this);
	}
	~Watchdog() {};
	int timeleft() {
		return _etime - time(0);
//		time_t curr;
//		time(&curr);
	}
private:
	pthread_t _tid; /* the thread identifier */
	pthread_attr_t _attr; /* set of thread attributes */	
	time_t _etime;
};
void *watchdog_runner(void* vp)
{
	Watchdog* pwd = static_cast<Watchdog*>(vp);
	cout << "watchdog initialised\n" << endl;
	while(pwd->timeleft() > 0) {
		//cout << "time left: " << pwd->timeleft() << endl;
		
	}
	pthread_exit(0);
}
#endif

/// main
int main(char argc,char* argv[])
{
//	Watchdog harakiri;
	openlog(::s_gpsname, LOG_PID, GPS_LOG_FACILITY);
	syslog(LOG_INFO,"started (ver.: %s built: %s %s)",VERSION,__DATE__,__TIME__);
	if(argc < 2)
		return end_usage("too few parameters");

	int cfgindex = 1;
	if(string(argv[1]) == string("-v")) {
		setlogmask(LOG_UPTO(LOG_DEBUG));
		g_getCfgNonConst().setVerbose(true);
		cfgindex = 2;
		if(argc < 3)
			return end_usage("parameter configfile missing");
	} else {
		setlogmask(LOG_UPTO(LOG_INFO));
	}
	syslog(LOG_DEBUG,"reading config: %s",argv[cfgindex]);
	ifstream cfgfile(argv[cfgindex]);
	if(!cfgfile) {
		syslog(LOG_ERR, "error opening config: %s, exiting",argv[cfgindex]);
		cerr << "Error opening config: " << argv[cfgindex] << ", exiting" << endl;
		//TODO says exiting, but doesnt
	}
//	cfgfile.setf(ios::skipws);
	// parse in config file
	ConfigReader reader(cfgfile);
	
	/// @param greydb is set if something went wrong
	///      during initialising but gps shouldnt exit     
	DB *greydb = 0;
	// try block for initialising phase
	// - parse config file
	// - connect to db
	try {
		g_getCfgNonConst().init(reader);
		//FIXME uses global Cfg object!!
		greydb = new DB; 
	// this should not exit, but continue with dunno
	// otherwise we see in the log files:
	//   "warning: problem talking to server private/policy: Success"
	// and a REJECT
	// This usually happens when the db is down
	} catch (exception &e) {
		greydb = 0;
		syslog(LOG_ERR, "error while initialising: %s, NO GREYLISTING WILL BE DONE",e.what());
	}
	// general try block for std::exception
	try { 
		Reader::AttributeSet attrSet;
		attrSet.insert(ATTR_CLIENT);
		attrSet.insert(ATTR_SENDER);
		attrSet.insert(ATTR_RECIPIENT);

		while(true) {
			try {
				Reader inputReader;
				int irSize = 0;
				do {
					inputReader.reset();
					irSize = inputReader.parse(attrSet,cin);
					if(irSize < 0) {
						if(greydb) {
							greydb->close();
							delete greydb;
							greydb = 0;
						}
						closelog();
						exit(0);
//						throw EndOfInputException("End of File, broken pipe?");
					}
					if( g_getCfg().isVerbose() ) {
						if( irSize != int(attrSet.size()))
							syslog(LOG_DEBUG,"ignoring garbage: %s", inputReader.getInput().c_str());
						else
							syslog(LOG_DEBUG,"input: %s", inputReader.getInput().c_str());
					}
				} while(irSize != int(attrSet.size()));
				// see if something went wrong during init
				if(!greydb) {
					Action act(Action::dunno); 
					continue; //skip greylisting
				}
				Triplet triplet(inputReader);
				DB::RecordStatus status = greydb->update(triplet);
				if(g_getCfg().isModeInit()) {
					Action act(Action::dunno);
				} else {
					switch(status) {
						default:
						case DB::registered:
							{ Action act(Action::dunno); }
							break;
						case DB::unknown:
						case DB::waiting:
							{ Action act(Action::defer); }
							break;

					}
				}
			} catch (EndOfInputException &e) {
				//throw it on
				throw e;
			} catch (exception &e) {
				syslog(LOG_WARNING, "exception caught: %s, continuing with \'dunno\'",e.what());
				Action act(Action::dunno);
			}
		} //while
	} catch (EndOfInputException &e) {
		syslog(LOG_INFO, "FATAL: %s, exiting",e.what());
		closelog();
		return 0;
	} catch (exception &e) {
		syslog(LOG_ERR, "unexcepted error: %s, exiting",e.what());
		Action act(Action::dunno);
		//cerr << "Error while initialising:\n\t" << e.what() << endl;
		closelog();
		return 0;
	}
	return 0;
}
